from typing import Union
from laktory.spark import SparkDataFrame
from laktory.spark import SparkColumn
from laktory.polars import PolarsLazyFrame
from laktory.polars import PolarsExpr

AnyDataFrame = Union[SparkDataFrame, PolarsLazyFrame]
AnyDataFrameColumn = Union[SparkColumn, PolarsExpr]



import os
import io
from pathlib import Path
import shutil
import uuid
import networkx as nx
import pandas as pd
import polars
from pyspark.sql import Window
import pyspark.sql.functions as F

from laktory import models
from laktory._testing import spark
from laktory._testing import Paths
from laktory._testing import df_brz

paths = Paths(__file__)

OPEN_FIGURES = False

testdir_path = Path(__file__).parent


def get_pl(clean_path=False):
    pl_path = testdir_path / "tmp" / "test_pipeline_polars" / str(uuid.uuid4())

    with open(os.path.join(paths.data, "pl-polars-local.yaml"), "r") as fp:
        data = fp.read()
        data = data.replace("{data_dir}", str(testdir_path / "data"))
        data = data.replace("{pl_dir}", str(pl_path))
        pl = models.Pipeline.model_validate_yaml(io.StringIO(data))

    if clean_path and os.path.exists(str(pl_path)):
        shutil.rmtree(str(pl_path))

    return pl, pl_path


gld_target = pd.DataFrame(
    {
        "symbol": ["AAPL", "GOOGL", "MSFT"],
        "max_price": [190.0, 138.0, 330.0],
        "min_price": [170.0, 129.0, 312.0],
        "mean_price": [177.0, 134.0, 320.0],
    }
)


def test_df_type():

    pl, _ = get_pl()

    # Check dataframe type assignment
    assert pl.dataframe_type == "POLARS"
    for node in pl.nodes:
        assert node.dataframe_type == "POLARS"
        assert node.source.dataframe_type == "POLARS"
        for s in node.get_sources():
            assert s.dataframe_type == "POLARS"


def test_execute():

    pl, pl_path = get_pl(clean_path=True)

    # Run
    pl.execute()

    # Test - Brz Stocks
    df = pl.nodes_dict["brz_stock_prices"].primary_sink.read().collect()
    assert df.columns == ["name", "description", "producer", "data", "_bronze_at"]
    assert df.height == 80

    # Test - Slv Meta
    df = pl.nodes_dict["slv_stock_meta"].output_df.collect()
    assert df.columns == ["symbol2", "currency", "first_traded"]
    assert df.height == 3

    # Test - Slv Stocks
    df = pl.nodes_dict["slv_stock_prices"].primary_sink.read().collect()
    assert df.columns == [
        "_bronze_at",
        "created_at",
        "symbol",
        "close",
        "currency",
        "first_traded",
        "_silver_at",
    ]
    assert df.height == 52

    # Test - Gold
    df = (
        pl.nodes_dict["gld_stock_prices"]
        .output_df.collect()
        .to_pandas()
        .round(0)
        .sort_values("symbol")
        .reset_index(drop=True)
    )
    assert len(df) == 3
    assert df.equals(gld_target)

    # Cleanup
    shutil.rmtree(pl_path)


def test_sql_join():

    # Get Pipeline
    pl, pl_path = get_pl(clean_path=True)

    # Update join
    node = pl.nodes_dict["slv_stock_prices"]
    t4 = node.transformer.nodes[-1]
    t4.sql_expr = """
    SELECT
        *
    FROM
        {df} as df
    LEFT JOIN
        {nodes.slv_stock_meta} as meta
    ON df.symbol = meta.symbol2
    ;
    """

    # Execute
    pl.execute(spark)

    # Test
    df = node.primary_sink.read().collect()
    assert df.columns == [
        "_bronze_at",
        "created_at",
        "symbol",
        "close",
        "symbol2",
        "currency",
        "first_traded",
        "_silver_at",
    ]

    # Cleanup
    shutil.rmtree(pl_path)


if __name__ == "__main__":
    test_df_type()
    test_execute()
    test_sql_join()



import os
from pydantic import model_validator
from typing import Union
from typing import Any
from settus import BaseSettings
from settus import Field
from settus import SettingsConfigDict


class Settings(BaseSettings):
    # model_config = SettingsConfigDict(
    #     keyvault_url="LAKTORY_KEYVAULT_URL"
    # )

    # CLI
    cli_raise_external_exceptions: bool = Field(
        False, alias="LAKTORY_CLI_RAISE_EXTERNAL_EXCEPTIONS"
    )

    # Azure
    lakehouse_sa_conn_str: Union[str, None] = Field(None, alias="LAKEHOUSE_SA_CONN_STR")

    # AWS
    aws_access_key_id: Union[str, None] = Field(None, alias="AWS_ACCESS_KEY_ID")
    aws_region: Union[str, None] = Field(None, alias="AWS_REGION")
    aws_secret_access_key: Union[str, None] = Field(None, alias="AWS_SECRET_ACCESS_KEY")

    # Databricks
    workspace_env: str = Field("dev", alias="LAKTORY_WORKSPACE_ENV")
    workspace_laktory_root: str = Field(
        "/.laktory/",
        alias="LAKTORY_WORKSPACE_LAKTORY_ROOT",
    )
    workspace_landing_root: str = Field("", alias="LAKTORY_WORKSPACE_LANDING_ROOT")

    # Paths
    laktory_root: str = Field("", alias="LAKTORY_ROOT")

    # Logging
    log_level: str = Field("INFO", alias="LAKTORY_LOG_LEVEL")

    @model_validator(mode="after")
    def update_landing_root(self) -> Any:
        if self.workspace_landing_root == "":
            self.workspace_landing_root = (
                f"/Volumes/{self.workspace_env}/sources/landing/"
            )

        return self

    @model_validator(mode="after")
    def update_laktory_root(self) -> Any:

        if self.laktory_root != "":
            return self

        # In Databricks
        # Could also use spark.conf.get("spark.databricks.cloudProvider") is not None
        if os.getenv("DATABRICKS_RUNTIME_VERSION"):
            self.laktory_root = "/laktory/"
        else:
            # Local execution
            self.laktory_root = "./"

        return self


settings = Settings()



import os
import shutil
import pytest
import uuid
from pathlib import Path
from pydantic_core._pydantic_core import ValidationError
from pyspark.sql import functions as F
from pyspark.sql import Window

from laktory import models
from laktory._testing import spark
from laktory._testing import Paths
from laktory._testing import df_brz
from laktory.exceptions import DataQualityCheckFailedError
from laktory.exceptions import DataQualityExpectationsNotSupported

paths = Paths(__file__)
testdir_path = Path(__file__).parent


def get_node():

    return models.PipelineNode(
        name="slv_stock_prices",
        source={
            "table_name": "brz_stock_prices",
            "mock_df": df_brz,
        },
        drop_source_columns=True,
        transformer={
            "nodes": [
                {
                    "with_columns": [
                        {
                            "name": "symbol",
                            "expr": "data.symbol",
                        },
                        {
                            "name": "close",
                            "expr": "data.close",
                            "type": "double",
                        },
                    ]
                },
            ],
        },
    )


def get_source(node_path):
    source_path = str(node_path / "brz_stock_prices")
    # w = Window.orderBy("data.created_at", "data.symbol")
    w = Window.orderBy("data.symbol", "data.created_at")
    source = df_brz.withColumn("index", F.row_number().over(w) - 1)
    return source, source_path


def test_warn():

    node = get_node()
    node.expectations = [
        models.DataQualityExpectation(
            name="max price",
            expr="close < 330",
            action="WARN",
        )
    ]
    node.execute()
    o = node.output_df.toPandas()
    assert node.checks[0].status == "FAIL"
    assert node.checks[0].rows_count == 80
    assert node.checks[0].fails_count == 8
    assert len(o) == 80


def test_drop():

    node = get_node()
    node.expectations = [
        models.DataQualityExpectation(
            name="max price",
            expr="close < 330",
            action="DROP",
        )
    ]
    node.execute()
    o = node.output_df.toPandas()
    assert node.checks[0].status == "FAIL"
    assert node.checks[0].rows_count == 80
    assert node.checks[0].fails_count == 8
    assert len(o) == 72


def test_quarantine():

    node = get_node()
    node.expectations = [
        models.DataQualityExpectation(
            name="max price",
            expr="close < 330",
            action="QUARANTINE",
        )
    ]
    node.execute()
    o = node.output_df.toPandas()
    q = node.quarantine_df.toPandas()
    assert node.checks[0].status == "FAIL"
    assert node.checks[0].rows_count == 80
    assert node.checks[0].fails_count == 8
    assert len(o) == 72
    assert len(q) == 8
    assert o["close"].max() < 330
    assert q["close"].min() >= 330


def test_fail():

    node = get_node()
    node.expectations = [
        models.DataQualityExpectation(
            name="max price",
            expr="close < 330",
            action="FAIL",
        )
    ]
    with pytest.raises(DataQualityCheckFailedError):
        node.execute()
    assert node.checks[0].status == "FAIL"
    assert node.checks[0].rows_count == 80
    assert node.checks[0].fails_count == 8


def test_aggregate():

    node = get_node()
    node.expectations = [
        models.DataQualityExpectation(
            name="rows count",
            expr="count(*) > 100",
            action="FAIL",
            type="AGGREGATE",
        )
    ]
    with pytest.raises(DataQualityCheckFailedError):
        node.execute()
    assert node.checks[0].status == "FAIL"
    assert node.checks[0].rows_count == 80
    assert node.checks[0].fails_count is None


def test_multi():

    node = get_node()
    node.expectations = [
        models.DataQualityExpectation(
            name="max price",
            expr="close < 330",
            action="QUARANTINE",
        ),
        models.DataQualityExpectation(
            name="not amazon",
            expr="symbol != 'AMZN'",
            action="DROP",
        ),
    ]
    node.execute()
    o = node.output_df.toPandas()
    q = node.quarantine_df.toPandas()
    assert node.checks[0].status == "FAIL"
    assert node.checks[0].rows_count == 80
    assert node.checks[0].fails_count == 8
    assert node.checks[1].status == "FAIL"
    assert node.checks[1].rows_count == 80
    assert node.checks[1].fails_count == 20
    assert len(o) == 52
    assert len(q) == 8
    assert o["close"].max() < 330
    assert q["close"].min() >= 330


def test_streaming_multi():

    node_path = (
        testdir_path / "tmp" / "test_pipeline_node_expectations" / str(uuid.uuid4())
    )
    checkpoint_path = node_path / "_checkpoint_expectations"

    # Cleanup
    if os.path.exists(str(node_path)):
        shutil.rmtree(str(node_path))

    # Create Stream Source
    source, source_path = get_source(node_path)

    # Insert a single row
    source.filter("index=0").write.format("delta").mode("OVERWRITE").save(source_path)

    # Get Node
    node = get_node()
    node.source = models.FileDataSource(
        path=source_path,
        format="DELTA",
        as_stream=True,
    )
    node.expectations = [
        models.DataQualityExpectation(
            name="max price",
            expr="close < 300",
            action="WARN",
        ),
        models.DataQualityExpectation(
            name="max price",
            expr="close < 330",
            action="QUARANTINE",
        ),
        models.DataQualityExpectation(
            name="not amazon",
            expr="symbol != 'AMZN'",
            action="DROP",
        ),
    ]
    node.expectations_checkpoint_location = checkpoint_path

    # Execute
    node.execute(spark=spark)

    # Test
    assert node.checks[0].status == "PASS"
    assert node.checks[0].rows_count == 1
    assert node.checks[0].fails_count == 0
    assert node.checks[1].status == "PASS"
    assert node.checks[1].rows_count == 1
    assert node.checks[1].fails_count == 0
    assert node.checks[2].status == "PASS"
    assert node.checks[2].rows_count == 1
    assert node.checks[2].fails_count == 0

    # Update source
    source.filter("index>0 AND index<40").write.format("delta").mode("append").save(
        source_path
    )
    node.execute(spark=spark)
    print(node.checks)
    assert node.checks[0].status == "PASS"
    assert node.checks[0].rows_count == 39
    assert node.checks[0].fails_count == 0
    assert node.checks[1].status == "PASS"
    assert node.checks[1].rows_count == 39
    assert node.checks[1].fails_count == 0
    assert node.checks[2].status == "FAIL"
    assert node.checks[2].rows_count == 39
    assert node.checks[2].fails_count == 20

    source.filter("index>=40").write.format("delta").mode("append").save(source_path)
    node.execute(spark=spark)
    print(node.checks)
    assert node.checks[0].status == "FAIL"
    assert node.checks[0].rows_count == 40
    assert node.checks[0].fails_count == 20
    assert node.checks[1].status == "FAIL"
    assert node.checks[1].rows_count == 40
    assert node.checks[1].fails_count == 8
    assert node.checks[2].status == "PASS"
    assert node.checks[2].rows_count == 40
    assert node.checks[2].fails_count == 0

    # Raise Exception
    source.filter("index>=40").write.format("delta").mode("append").save(source_path)
    node.expectations[0].action = "FAIL"
    # node.execute(spark=spark)
    # # TODO: Find how to capture exception
    # # with pytest.raises(Exception):
    # #     node.execute(spark=spark)


def test_expectations_invalid():

    with pytest.raises(DataQualityExpectationsNotSupported):
        node = models.PipelineNode(
            name="slv_stock_prices",
            source={
                "path": "some_path",
                "format": "DELTA",
                "as_stream": True,
            },
            expectations=[
                {
                    "name": "max price pass",
                    "expr": "close < 300",
                    "action": "WARN",
                },
                {
                    "name": "max price drop",
                    "expr": "count(*) > 20",
                    "type": "AGGREGATE",
                },
            ],
            expectations_checkpoint_location="some_path",
        )

    with pytest.raises(DataQualityExpectationsNotSupported):
        node = models.PipelineNode(
            name="slv_stock_prices",
            source={
                "path": "some_path",
                "format": "DELTA",
                "as_stream": True,
            },
            expectations=[
                {
                    "name": "max price pass",
                    "expr": "close < 300",
                    "action": "WARN",
                    "tolerance": {"abs": 20},
                },
            ],
            expectations_checkpoint_location="some_path",
        )

    # with pytest.warns(UserWarning):
    #     node = models.PipelineNode(
    #         name="slv_stock_prices",
    #         source={
    #             "path": "some_path",
    #             "format": "DELTA",
    #             "as_stream": True,
    #         },
    #         expectations=[
    #             {
    #                 "name": "max price pass",
    #                 "expr": "F.('close') < 300",
    #             },
    #         ],
    #     )


if __name__ == "__main__":
    test_warn()
    test_drop()
    test_quarantine()
    test_fail()
    test_aggregate()
    test_multi()
    test_streaming_multi()
    test_expectations_invalid()



import pytest
from pyspark.sql import functions as F

from laktory import models
from laktory.exceptions import DataQualityCheckFailedError
from laktory._testing import Paths
from laktory._testing import df_slv as df

paths = Paths(__file__)


def test_expectations_abs():

    # Spark Expression - WARN
    dqe = models.DataQualityExpectation(
        name="price less than 300", action="WARN", expr="F.col('close') < 300"
    )
    check = dqe.run_check(df)
    assert check.rows_count == 80
    assert check.fails_count == 20
    assert check.failure_rate == 0.25
    assert check.status == "FAIL"
    assert str(dqe.pass_filter) == str(F.col("close") < 300)
    assert str(dqe.fail_filter) == str(~(F.col("close") < 300))
    assert dqe.keep_filter is None
    assert dqe.quarantine_filter is None

    # Spark Expression - QUARANTINE
    dqe = models.DataQualityExpectation(
        name="price less than 300", action="DROP", expr="F.col('close') < 300"
    )
    check = dqe.run_check(df)
    assert check.rows_count == 80
    assert check.fails_count == 20
    assert check.failure_rate == 0.25
    assert check.status == "FAIL"
    assert str(dqe.pass_filter) == str(F.col("close") < 300)
    assert str(dqe.fail_filter) == str(~(F.col("close") < 300))
    assert str(dqe.keep_filter) == str(F.col("close") < 300)
    assert dqe.quarantine_filter is None

    # Spark Expression - QUARANTINE
    dqe = models.DataQualityExpectation(
        name="price less than 300", action="QUARANTINE", expr="F.col('close') < 300"
    )
    check = dqe.run_check(df)
    assert check.rows_count == 80
    assert check.fails_count == 20
    assert check.failure_rate == 0.25
    assert check.status == "FAIL"
    assert str(dqe.pass_filter) == str(F.col("close") < 300)
    assert str(dqe.fail_filter) == str(~(F.col("close") < 300))
    assert str(dqe.keep_filter) == str(F.col("close") < 300)
    assert str(dqe.quarantine_filter) == str(~(F.col("close") < 300))

    # SQL Expression
    dqe = models.DataQualityExpectation(
        name="price less than 300", action="WARN", expr="close < 300"
    )
    check = dqe.run_check(df)
    assert check.rows_count == 80
    assert check.fails_count == 20
    assert check.failure_rate == 0.25
    assert check.status == "FAIL"
    assert dqe.keep_filter is None
    assert dqe.quarantine_filter is None


def test_expectations_rel():

    dqe = models.DataQualityExpectation(
        name="price higher than 10",
        action="WARN",
        expr="close > 127",
        tolerance={"rel": 0.05},
    )
    check = dqe.run_check(df)
    assert check.rows_count == 80
    assert check.fails_count == 3
    assert check.failure_rate == 0.0375
    assert check.status == "PASS"
    assert dqe.keep_filter is None
    assert dqe.quarantine_filter is None


def test_expectations_agg():

    dqe = models.DataQualityExpectation(
        name="rows count",
        expr="COUNT(*) > 50",
        type="AGGREGATE",
    )
    check = dqe.run_check(df)
    assert check.rows_count == 80
    assert check.fails_count is None
    assert check.failure_rate is None
    assert check.status == "PASS"
    assert dqe.keep_filter is None
    assert dqe.quarantine_filter is None

    dqe = models.DataQualityExpectation(
        name="rows count",
        expr="F.count('*') > 90",
        type="AGGREGATE",
    )
    check = dqe.run_check(df)
    assert check.rows_count == 80
    assert check.fails_count is None
    assert check.failure_rate is None
    assert check.status == "FAIL"
    assert dqe.keep_filter is None
    assert dqe.quarantine_filter is None


def test_expectations_empty():

    # Spark Expression
    dqe = models.DataQualityExpectation(
        name="price less than 300", action="WARN", expr="F.col('close') < 300"
    )
    check = dqe.run_check(df.filter("close < 0"))
    assert check.rows_count == 0
    assert check.fails_count == 0
    assert check.failure_rate == 0
    assert check.status == "PASS"
    assert dqe.keep_filter is None
    assert dqe.quarantine_filter is None


def test_expectations_exceptions_warnings():

    # No Failure
    dqe = models.DataQualityExpectation(
        name="price less than 900", action="FAIL", expr="F.col('close') < 900"
    )
    dqe.run_check(df, raise_or_warn=True)

    # Do Not Raise
    dqe = models.DataQualityExpectation(
        name="price less than 300", action="WARN", expr="F.col('close') < 300"
    )
    dqe.run_check(df, raise_or_warn=False)

    # Warn
    dqe = models.DataQualityExpectation(
        name="price less than 300", action="WARN", expr="F.col('close') < 300"
    )
    with pytest.warns(UserWarning):
        dqe.run_check(df, raise_or_warn=True)

    # Raise Exception
    dqe = models.DataQualityExpectation(
        name="price less than 300", action="FAIL", expr="F.col('close') < 300"
    )
    with pytest.raises(DataQualityCheckFailedError):
        dqe.run_check(df, raise_or_warn=True)


if __name__ == "__main__":
    test_expectations_abs()
    test_expectations_rel()
    test_expectations_agg()
    test_expectations_empty()
    test_expectations_exceptions_warnings()



import os
from databricks.connect import DatabricksSession

# Setup
spark = DatabricksSession.builder.clusterId("0501-011833-vcux5w7j").getOrCreate()

# Execute
df = spark.readStream.table("brz_stock_prices_job")


def update_metrics(batch_df, batch_id):
    size = batch_df.count()
    print(f"Batch size: {size}")


writer = df.writeStream.foreachBatch(update_metrics).start()

writer.awaitTermination()



import os
import io
from pathlib import Path
from laktory import models
from laktory._testing import Paths

paths = Paths(__file__)
testdir_path = Path(__file__).parent


def get_pl(extra=None):
    with open(os.path.join(paths.data, "pl-spark-local.yaml"), "r") as fp:
        data = fp.read()
        data = data.replace("{data_dir}", "data")
        data = data.replace("{pl_dir}", "")
        if extra is not None:
            data += extra
        pl = models.Pipeline.model_validate_yaml(io.StringIO(data))
    return pl


# Job
pl_job = get_pl(
    """
name: pl-spark-job
orchestrator: DATABRICKS_JOB
databricks_job:
  name: job-pl-stock-prices
  laktory_version: 0.3.0
  clusters:
    - name: node-cluster
      spark_version: 14.0.x-scala2.12
      node_type_id: Standard_DS3_v2
"""
)

# DLT
pl_dlt = get_pl(
    """
name: pl-spark-dlt
orchestrator: DLT
dlt:
  catalog: dev
  target: sandbox
  access_controls:
  - group_name: account users
    permission_level: CAN_VIEW
  options:
    provider: ${resources.databricks2}
options:
  provider: ${resources.databricks1}
"""
)


def test_pipeline_job():

    # Test job
    job = pl_job.databricks_job
    data = job.model_dump(exclude_unset=True)
    print(data)
    assert data == {
        "clusters": [
            {
                "name": "node-cluster",
                "node_type_id": "Standard_DS3_v2",
                "spark_version": "14.0.x-scala2.12",
            }
        ],
        "name": "job-pl-stock-prices",
        "parameters": [
            {"default": "false", "name": "full_refresh"},
            {"default": "pl-spark-job", "name": "pipeline_name"},
        ],
        "tasks": [
            {
                "depends_ons": [],
                "job_cluster_key": "node-cluster",
                "notebook_task": {
                    "notebook_path": "/.laktory/jobs/job_laktory_pl.py",
                    "base_parameters": {"node_name": "brz_stock_meta"},
                },
                "task_key": "node-brz_stock_meta",
            },
            {
                "depends_ons": [],
                "job_cluster_key": "node-cluster",
                "notebook_task": {
                    "notebook_path": "/.laktory/jobs/job_laktory_pl.py",
                    "base_parameters": {"node_name": "brz_stock_prices"},
                },
                "task_key": "node-brz_stock_prices",
            },
            {
                "depends_ons": [{"task_key": "node-slv_stock_prices"}],
                "job_cluster_key": "node-cluster",
                "notebook_task": {
                    "notebook_path": "/.laktory/jobs/job_laktory_pl.py",
                    "base_parameters": {"node_name": "gld_stock_prices"},
                },
                "task_key": "node-gld_stock_prices",
            },
            {
                "depends_ons": [{"task_key": "node-brz_stock_meta"}],
                "job_cluster_key": "node-cluster",
                "notebook_task": {
                    "notebook_path": "/.laktory/jobs/job_laktory_pl.py",
                    "base_parameters": {"node_name": "slv_stock_meta"},
                },
                "task_key": "node-slv_stock_meta",
            },
            {
                "depends_ons": [
                    {"task_key": "node-brz_stock_prices"},
                    {"task_key": "node-slv_stock_meta"},
                ],
                "job_cluster_key": "node-cluster",
                "notebook_task": {
                    "notebook_path": "/.laktory/jobs/job_laktory_pl.py",
                    "base_parameters": {"node_name": "slv_stock_prices"},
                },
                "task_key": "node-slv_stock_prices",
            },
        ],
        "laktory_version": "0.3.0",
    }

    # Test resources
    resources = pl_job.core_resources
    assert len(resources) == 3


def test_pipeline_dlt():

    # Test Sink as Source
    sink_source = pl_dlt.nodes[1].source.node.primary_sink.as_source(
        as_stream=pl_dlt.nodes[1].source.as_stream
    )
    data = sink_source.model_dump()
    print(data)
    assert data == {
        "as_stream": True,
        "broadcast": False,
        "dataframe_type": "SPARK",
        "drops": None,
        "filter": None,
        "limit": None,
        "renames": None,
        "sample": None,
        "selects": None,
        "watermark": None,
        "format": "DELTA",
        "path": "/brz_stock_prices",
        "read_options": {},
        "schema_definition": None,
        "schema_location": None,
    }

    data = pl_dlt.dlt.model_dump()
    print(data)
    assert data == {
        "access_controls": [
            {
                "group_name": "account users",
                "permission_level": "CAN_VIEW",
                "service_principal_name": None,
                "user_name": None,
            }
        ],
        "allow_duplicate_names": None,
        "catalog": "dev",
        "channel": "PREVIEW",
        "clusters": [],
        "configuration": {},
        "continuous": None,
        "development": None,
        "edition": None,
        "libraries": None,
        "name": "pl-spark-dlt",
        "notifications": [],
        "photon": None,
        "serverless": None,
        "storage": None,
        "target": "sandbox",
    }

    # Test resources
    resources = pl_dlt.core_resources
    assert len(resources) == 4

    wsf = resources[0]
    wsfp = resources[1]
    dlt = resources[2]
    dltp = resources[3]

    assert isinstance(wsf, models.resources.databricks.WorkspaceFile)
    assert isinstance(wsfp, models.resources.databricks.Permissions)
    assert isinstance(dlt, models.resources.databricks.DLTPipeline)
    assert isinstance(dltp, models.resources.databricks.Permissions)

    assert dlt.resource_name == "pl-spark-dlt"
    assert dltp.resource_name == "permissions-pl-spark-dlt"
    assert wsf.resource_name == "workspace-file-laktory-pipelines-pl-spark-dlt-json"
    assert (
        wsfp.resource_name
        == "permissions-workspace-file-laktory-pipelines-pl-spark-dlt-json"
    )

    assert dlt.options.provider == "${resources.databricks2}"
    assert dltp.options.provider == "${resources.databricks2}"
    assert wsf.options.provider == "${resources.databricks1}"
    assert wsfp.options.provider == "${resources.databricks1}"

    assert dlt.options.depends_on == []
    assert dltp.options.depends_on == ["${resources.pl-spark-dlt}"]
    assert wsf.options.depends_on == []
    assert wsfp.options.depends_on == [
        "${resources.workspace-file-laktory-pipelines-pl-spark-dlt-json}"
    ]


if __name__ == "__main__":
    test_pipeline_job()
    test_pipeline_dlt()



import os
import shutil
import uuid
from pathlib import Path
from py import path as pypath

from laktory import app
from laktory import settings
from laktory import models
from laktory._testing import Paths
from typer.testing import CliRunner

runner = CliRunner()
settings.cli_raise_external_exceptions = True
paths = Paths(__file__)


def _read_stack(template, backend):

    dirpath = pypath.local(
        f"{paths.tmp}/quickstart_{template}_{backend}_{str(uuid.uuid4())}"
    )
    stack_filepath = dirpath / "stack.yaml"

    # Change dir
    os.mkdir(dirpath)

    # Copy Stack
    with dirpath.as_cwd():
        _ = runner.invoke(
            app,
            ["quickstart", "--template", template, "--backend", backend],
        )

    # Read Stack
    with open(stack_filepath) as fp:
        stack = models.Stack.model_validate_yaml(fp)
    data = stack.model_dump()

    # Count resources
    resources_count = 0
    for k in data["resources"]:
        if k == "providers":
            continue
        resources_count += len(data["resources"][k].keys())

    if template == "workspace":
        assert resources_count == 7
        assert (
            "directory-laktory-dashboards"
            in data["resources"]["databricks_directories"].keys()
        )
        assert (
            "secret-scope-laktory"
            in data["resources"]["databricks_secretscopes"].keys()
        )
        assert "warehouse-laktory" in data["resources"]["databricks_warehouses"].keys()
        pass
    elif template == "workflows":
        assert resources_count == 7
        assert (
            "dbfs-file-stock-prices" in data["resources"]["databricks_dbfsfiles"].keys()
        )
        assert (
            "notebook-job-laktory-pl"
            in data["resources"]["databricks_notebooks"].keys()
        )
        assert (
            "notebook-dlt-laktory-pl"
            in data["resources"]["databricks_notebooks"].keys()
        )
        assert "job-hello" in data["resources"]["databricks_jobs"].keys()
        assert "pl-stocks-job" in data["resources"]["pipelines"].keys()
        assert "pl-stocks-dlt" in data["resources"]["pipelines"].keys()
    elif template == "unity-catalog":
        assert resources_count == 8
        assert "group-laktory-friends" in data["resources"]["databricks_groups"].keys()
        assert "user-kubic-okube-ai" in data["resources"]["databricks_users"].keys()
        assert "catalog-lqs-prd" in data["resources"]["databricks_catalogs"].keys()

    else:
        raise ValueError()

    # Cleanup
    shutil.rmtree(dirpath)


def _preview_stack(template, backend, env):

    dirpath = pypath.local(
        f"{paths.tmp}/quickstart_{template}_{backend}_{str(uuid.uuid4())}"
    )
    stack_filepath = dirpath / "stack.yaml"

    # Change dir
    os.mkdir(dirpath)

    # Copy Stack
    with dirpath.as_cwd():
        _ = runner.invoke(
            app,
            ["quickstart", "--template", template, "--backend", backend],
        )

    if backend == "terraform":
        # Ideally, we would run `laktory init`, but the runner does not seem to handle running multiple commands
        with open(stack_filepath, "r") as fp:
            pstack = models.Stack.model_validate_yaml(fp).to_terraform(env_name=env)
            pstack.init(flags=["-migrate-state", "-upgrade"])

    # Preview Stack
    cmds = ["preview"]
    if env:
        cmds += ["--env", env]
    _ = runner.invoke(app, cmds)

    # Cleanup
    shutil.rmtree(dirpath)


def _deploy_stack(template, backend, env):

    dirpath = pypath.local(
        f"{paths.tmp}/quickstart_{template}_{backend}_{str(uuid.uuid4())}"
    )
    stack_filepath = dirpath / "stack.yaml"

    # Change dir
    os.mkdir(dirpath)

    # Copy Stack
    with dirpath.as_cwd():
        _ = runner.invoke(
            app,
            ["quickstart", "--template", template, "--backend", backend],
        )

    if backend == "terraform":
        # Ideally, we would run `laktory init`, but the runner does not seem to handle running multiple commands
        with open(stack_filepath, "r") as fp:
            pstack = models.Stack.model_validate_yaml(fp).to_terraform(env_name=env)
            pstack.init(flags=["-migrate-state", "-upgrade"])

    # Deploy Stack
    cmds = ["deploy"]
    if env:
        cmds += ["-e", env]
    if backend == "pulumi":
        cmds += ["--pulumi-options", "--yes"]
    elif backend == "terraform":
        cmds += ["--terraform-options", "--auto-approve"]
    _ = runner.invoke(app, cmds)

    # Cleanup
    shutil.rmtree(dirpath)


def test_read_quickstart_stacks():
    for template, backend in [
        ("workspace", "terraform"),
        ("workspace", "pulumi"),
        ("workflows", "terraform"),
        ("workflows", "pulumi"),
        ("unity-catalog", "terraform"),
        ("unity-catalog", "pulumi"),
    ]:
        _read_stack(template, backend)


def test_preview_quickstart_stacks():
    for template, backend, env in [
        ("workspace", "terraform", "dev"),
        ("workspace", "pulumi", "dev"),
        ("workflows", "terraform", "dev"),
        ("workflows", "pulumi", "dev"),
        ("unity-catalog", "terraform", None),
        ("unity-catalog", "pulumi", None),
    ]:
        _preview_stack(template, backend, env)


def atest_deploy_quickstart_stacks():
    for template, backend, env in [
        ("workspace", "terraform", "dev"),
        ("workspace", "pulumi", "dev"),
        ("workflows", "terraform", "dev"),
        ("workflows", "pulumi", "dev"),
        ("unity-catalog", "terraform"),
        ("unity-catalog", "pulumi"),
    ]:
        _deploy_stack(template, backend, env)


def test_quickstart_localpipeline():

    dirpath = pypath.local(f"{paths.tmp}/quickstart_local_pipeline_{str(uuid.uuid4())}")
    # stack_filepath = dirpath / "stack.yaml"

    # Change dir
    os.mkdir(dirpath)

    with dirpath.as_cwd():

        # Run Quickstart
        _ = runner.invoke(
            app,
            ["quickstart", "--template", "local-pipeline"],
        )

        # Run Scripts
        for filename in [
            "00_explore_pipeline.py",
            "01_execute_node_bronze.py",
            "02_execute_node_silver.py",
            "03_execute_pipeline.py",
            "04_code_pipeline.py",
        ]:
            with open(filename, "r") as f:
                print(f"----- Executing {filename}")
                code = f.read()
                exec(code)
                print("")
                print(f"----- Execution completed\n\n")

    # Cleanup
    shutil.rmtree(dirpath)


if __name__ == "__main__":
    test_read_quickstart_stacks()
    test_preview_quickstart_stacks()
    atest_deploy_quickstart_stacks()
    test_quickstart_localpipeline()



import json
from laktory.models.resources.databricks import ClusterPolicy

definition = {
    "dbus_per_hour": {
        "type": "range",
        "maxValue": 10,
    },
    "autotermination_minutes": {"type": "fixed", "value": 30, "hidden": True},
    "custom_tags.team": {
        "type": "fixed",
        "value": "okube",
    },
}

cluster_policy = ClusterPolicy(
    name="okube",
    definition=json.dumps(definition),
    libraries=[
        {
            "pypi": {
                "package": "laktory==0.5.0",
            }
        }
    ],
    access_controls=[{"permission_level": "CAN_USE", "group_name": "account users"}],
)


def test_policy_cluster():
    assert cluster_policy.name == "okube"
    assert cluster_policy.definition == json.dumps(definition)


def test_policy_cluster_as_dict():

    cp = ClusterPolicy(
        name="okube",
        definition=definition,
        libraries=[
            {
                "pypi": {
                    "package": "laktory==0.5.0",
                }
            }
        ],
        access_controls=[
            {"permission_level": "CAN_USE", "group_name": "account users"}
        ],
    )

    assert cluster_policy.name == "okube"
    assert cp.definition == json.dumps(definition)


if __name__ == "__main__":
    test_policy_cluster()



from typing import Union
import os.path
import json

from typing import Literal
from typing import Any
from pydantic import model_validator
from pydantic import Field
from pydantic import ConfigDict

from laktory.models.datasources.basedatasource import BaseDataSource
from laktory.spark import SparkDataFrame
from laktory.polars import PolarsLazyFrame
from laktory._logger import get_logger

logger = get_logger(__name__)


class FileDataSource(BaseDataSource):
    """
    Data source using disk files, such as data events (json/csv) and
    dataframe parquets. It is generally used in the context of a data pipeline.

    Attributes
    ----------
    format:
        Format of the data files
    read_options:
        Other options passed to `spark.read.options`
    schema:
        Target schema specified as a list of columns, as a dict or a json
        serialization. Only used when reading data from non-strongly typed
        files such as JSON or csv files.
    schema_location:
        Path for schema inference when reading data as a stream. If `None`,
        parent directory of `path` is used.

    Examples
    ---------
    ```python
    from laktory import models

    source = models.FileDataSource(
        path="/Volumes/sources/landing/events/yahoo-finance/stock_price",
        format="JSON",
        as_stream=False,
    )
    # df = source.read(spark)

    # With Explicit Schema
    source = models.FileDataSource(
        path="/Volumes/sources/landing/events/yahoo-finance/stock_price",
        format="JSON",
        as_stream=False,
        schema=[
            {"name": "description", "type": "string", "nullable": True},
            {"name": "close", "type": "double", "nullable": False},
        ],
    )
    # df = source.read(spark)
    ```
    """

    model_config = ConfigDict(populate_by_name=True)

    format: Literal[
        "CSV", "PARQUET", "DELTA", "JSON", "NDJSON", "JSONL", "EXCEL", "BINARYFILE"
    ] = "JSONL"
    path: str
    read_options: dict[str, Any] = {}
    schema_definition: Union[str, dict, list] = Field(None, validation_alias="schema")
    schema_location: str = None

    @model_validator(mode="after")
    def options(self) -> Any:

        if self.dataframe_type == "SPARK":
            if self.format in [
                "EXCEL",
            ]:
                raise ValueError(f"'{self.format}' format is not supported with Spark")

        elif self.dataframe_type == "POLARS":
            if self.format in [
                "BINARYFILE",
            ]:
                raise ValueError(f"'{self.format}' format is not supported with Polars")

        return self

    # ----------------------------------------------------------------------- #
    # Properties                                                              #
    # ----------------------------------------------------------------------- #

    @property
    def _id(self):
        return str(self.path)

    @property
    def _schema(self):
        schema = self.schema_definition
        if schema is None:
            return schema
        if isinstance(schema, list):
            schema = {"fields": schema, "type": "struct"}

        if isinstance(schema, str) and '"fields":[' in schema.replace("'", '"').replace(
            " ", ""
        ):
            schema = json.loads(schema)

        # DDL format
        if isinstance(schema, str):
            pass

        # Spark Struct format
        elif isinstance(schema, dict):
            import pyspark.sql.types as T

            schema = T.StructType.fromJson(schema)

        return schema

    # ----------------------------------------------------------------------- #
    # Readers                                                                 #
    # ----------------------------------------------------------------------- #

    def _read_spark(self, spark) -> SparkDataFrame:

        _options = {}
        _mode = "stream"

        # JSON
        _format = self.format
        if self.format in ["NDJSON", "JSONL"]:
            _format = "JSON"
            _options["multiline"] = False
        elif self.format == "JSON":
            _format = "JSON"
            _options["multiline"] = True

        # CSV
        if self.format == "CSV":
            _options["header"] = True

        if self.as_stream:
            _mode = "stream"

            if _format == "DELTA":
                reader = spark.readStream.format(_format)

            else:
                reader = spark.readStream.format("cloudFiles")
                _options["cloudFiles.format"] = _format

                if self._schema:
                    reader = reader.schema(self._schema)
                else:
                    schema_location = self.schema_location
                    if schema_location is None:
                        schema_location = os.path.dirname(self.path)
                    _options["cloudFiles.schemaLocation"] = schema_location

            if self._schema is None:
                _options["cloudFiles.inferColumnTypes"] = True
                _options["cloudFiles.schemaEvolutionMode"] = "addNewColumns"

        else:
            _mode = "static"
            reader = spark.read.format(_format)
            if self._schema:
                reader = reader.schema(self._schema)

        # User Options
        _options["mergeSchema"] = True
        _options["recursiveFileLookup"] = True
        if self.read_options:
            for k, v in self.read_options.items():
                _options[k] = v

        reader = reader.options(**_options)

        # Load
        logger.info(f"Reading {self._id} as {_mode} and options {_options}")
        if self._schema:
            schema_str = self._schema
            if hasattr(schema_str, "simpleString"):
                schema_str = schema_str.simpleString()
            else:
                schema_str = str(schema_str)
            logger.info(f"Expected schema: {schema_str}")
        df = reader.load(self.path)

        return df

    def _read_polars(self) -> PolarsLazyFrame:

        import polars as pl

        if self.as_stream:
            raise ValueError(
                "Streaming read not supported with Pandas DataFrame. Please switch to Spark"
            )

        logger.info(f"Reading {self._id} as static")

        if self.format.lower() == "csv":
            df = pl.scan_csv(self.path, **self.read_options)

        elif self.format.lower() == "delta":
            df = pl.scan_delta(self.path, **self.read_options)

        elif self.format.lower() == "excel":
            df = pl.read_excel(self.path, **self.read_options)

        elif self.format.lower() == "json":
            df = pl.read_json(self.path, **self.read_options)

        elif self.format.lower() in ["jsonl", "ndjson"]:
            df = pl.scan_ndjson(self.path, **self.read_options)

        elif self.format.lower() == "parquet":
            df = pl.scan_parquet(self.path, **self.read_options)

        else:
            raise ValueError(f"Format '{self.format}' is not supported.")

        if isinstance(df, pl.DataFrame):
            df = df.lazy()

        return df



from __future__ import annotations
from pydantic import field_validator
from pydantic import model_validator
from typing import Any
from typing import Union
from typing import Callable
from typing import Literal
from typing import TYPE_CHECKING
import abc
import re

from laktory._logger import get_logger
from laktory.constants import SUPPORTED_DATATYPES
from laktory.constants import DEFAULT_DFTYPE
from laktory.models.basemodel import BaseModel
from laktory.models.dataframecolumnexpression import DataFrameColumnExpression
from laktory.polars import PolarsDataFrame
from laktory.polars import PolarsExpr
from laktory.types import AnyDataFrame

if TYPE_CHECKING:
    from laktory.models.datasources.basedatasource import BaseDataSource
    from laktory.models.datasources.pipelinenodedatasource import PipelineNodeDataSource


logger = get_logger(__name__)


# --------------------------------------------------------------------------- #
# Helper Classes                                                              #
# --------------------------------------------------------------------------- #


class BaseChainNodeFuncArg(BaseModel):
    value: Union[Any]

    @field_validator("value")
    def value_to_data_source(cls, v: Any) -> Any:
        """
        Data source can't be set as an expected value type as it would create
        a circular dependency. Instead, we check at validation if the value
        could be instantiated as a DataSource object.
        """
        from laktory.models.datasources import classes

        for c in classes:
            try:
                v = c(**v)
                break
            except:
                pass

        return v

    @abc.abstractmethod
    def eval(self):
        raise NotImplementedError()

    def signature(self):
        from laktory.models.datasources import PipelineNodeDataSource
        from laktory.models.datasources import FileDataSource
        from laktory.models.datasources import TableDataSource

        if isinstance(self.value, PipelineNodeDataSource):
            return f"node.{self.value.node_name}"
        elif isinstance(self.value, FileDataSource):
            return f"file {self.value.path}"
        elif isinstance(self.value, TableDataSource):
            return f"table {self.value.full_name}"
        return str(self.value)


class ChainNodeColumn(BaseModel):
    expr: Union[str, DataFrameColumnExpression]
    name: str
    type: Union[str, None] = "string"
    unit: Union[str, None] = None

    @model_validator(mode="after")
    def parse_expr(self) -> Any:
        if isinstance(self.expr, str):
            self.expr = DataFrameColumnExpression(value=self.expr)
        return self

    @field_validator("type")
    def check_type(cls, v: str) -> str:

        if v is None or "<" in v:
            return v
        else:
            if v not in SUPPORTED_DATATYPES:
                raise ValueError(
                    f"Type {v} is not supported. Select one of {SUPPORTED_DATATYPES}"
                )
        return v

    def eval(self, udfs=None, dataframe_type=DEFAULT_DFTYPE):
        return self.expr.eval(udfs=udfs, dataframe_type=dataframe_type)


class BaseChainNodeSQLExpr(BaseModel):
    """
    Chain node SQL expression

    Attributes
    ----------
    expr:
        SQL expression
    """

    expr: str
    _node_data_sources: list[PipelineNodeDataSource] = None

    def parsed_expr(self, df_id="df"):
        return self.expr

    @property
    def node_data_sources(self) -> list[PipelineNodeDataSource]:

        if self.expr is None:
            return []

        if self._node_data_sources is None:

            from laktory.models.datasources.pipelinenodedatasource import (
                PipelineNodeDataSource,
            )

            sources = []

            pattern = r"\{nodes\.(.*?)\}"
            matches = re.findall(pattern, self.expr)
            for m in matches:
                sources += [PipelineNodeDataSource(node_name=m)]

            self._node_data_sources = sources

        return self._node_data_sources

    def eval(self, df, chain_node=None):
        raise NotImplementedError()


# --------------------------------------------------------------------------- #
# Main Class                                                                  #
# --------------------------------------------------------------------------- #


class BaseChainNode(BaseModel):

    dataframe_type: Literal["SPARK", "POLARS", None] = "SPARK"
    func_args: list[Union[Any]] = []
    func_kwargs: dict[str, Union[Any]] = {}
    func_name: Union[str, None] = None
    sql_expr: Union[str, None] = None
    with_column: Union[ChainNodeColumn, None] = None
    with_columns: Union[list[ChainNodeColumn], None] = []
    _parent: "BaseChain" = None
    _parsed_func_args: list = None
    _parsed_func_kwargs: dict = None
    _parsed_sql_expr: BaseChainNodeSQLExpr = None

    @model_validator(mode="after")
    def selected_flow(self) -> Any:

        if len(self._with_columns) > 0:
            if self.func_name:
                raise ValueError(
                    "`func_name` should not be set when using `with_column`"
                )
            if self.sql_expr:
                raise ValueError(
                    "`sql_expr` should not be set when using `with_column`"
                )
        else:
            if self.func_name and self.sql_expr:
                raise ValueError(
                    "Only one of `func_name` and `sql_expr` should be set."
                )
            if not (self.func_name or self.sql_expr):
                raise ValueError(
                    "Either `func_name`, `sql_expr` or `with_column` should be set when using"
                )

        return self

    @property
    def _with_columns(self) -> list[ChainNodeColumn]:
        with_columns = [c for c in self.with_columns]
        if self.with_column:
            with_columns += [self.with_column]
        return with_columns

    @property
    def is_column(self):
        return len(self._with_columns) > 0

    @property
    def id(self):
        if self.is_column:
            return "-".join([c.name for c in self.with_columns])
        return "df"

    @property
    def user_dftype(self):
        if "dataframe_type" in self.model_fields_set:
            return self.dataframe_type
        return None

    # ----------------------------------------------------------------------- #
    # Class Methods                                                           #
    # ----------------------------------------------------------------------- #

    def get_sources(self, cls=None) -> list[BaseDataSource]:
        """Get all sources feeding the Chain Node"""

        from laktory.models.datasources.basedatasource import BaseDataSource

        if cls is None:
            cls = BaseDataSource

        sources = []
        for a in self.parsed_func_args:
            if isinstance(a.value, cls):
                sources += [a.value]
        for a in self.parsed_func_kwargs.values():
            if isinstance(a.value, cls):
                sources += [a.value]

        if self.sql_expr:
            sources += self.parsed_sql_expr.node_data_sources

        return sources

    @abc.abstractmethod
    def execute(
        self,
        df: AnyDataFrame,
        udfs: list[Callable[[...], Union[PolarsExpr, PolarsDataFrame]]] = None,
        return_col: bool = False,
    ) -> Union[AnyDataFrame]:
        raise NotImplementedError()
